# -*- coding: utf-8 -*-

'''
LANdmark-based Transferable Subspace Analysis
'''

__version__ = '0.4.0'


from ._select_learning_samples import select_learning_samples
from ._select_landmarks import select_landmarks
from ._subspace_analysis import subspace_analysis
from ._transfer_labels import transfer_labels









